from __future__ import annotations  # noqa: I2041

from typing import Callable, List, TYPE_CHECKING

if TYPE_CHECKING:
    from datetimeparser.enums import ConstantOption  # noqa: I2041


class Printable:
    FIELDS = []

    def __str__(self):
        return f'<{str(self.__class__)[7:-1]} {" ".join("%s=%s" % (field, getattr(self, field)) for field in self.FIELDS if getattr(self, field) is not None)}>'

    def __repr__(self):
        return self.__str__()


class Concatenable(Printable):
    @classmethod
    def _concatenate(cls, new, o1, o2):
        for field in cls.FIELDS:
            d1 = getattr(o1, field)
            d2 = getattr(o2, field)

            if d1 == 0 and d2 == 0:
                continue

            if d1 != 0 and d2 != 0:
                setattr(new, field, d2)

            if d1 == 0:
                setattr(new, field, d2)
            elif d2 == 0:
                setattr(new, field, d1)

        return new


class AbsoluteDateTime(Concatenable):
    FIELDS = ["year", "month", "day", "hour", "minute", "second"]

    def __init__(self, year=0, month=0, day=0, hour=0, minute=0, second=0):
        self.year = year
        self.month = month
        self.day = day
        self.hour = hour
        self.minute = minute
        self.second = second

    @classmethod
    def concatenate(cls, o1: 'AbsoluteDateTime', o2: 'AbsoluteDateTime') -> 'AbsoluteDateTime':
        return cls._concatenate(AbsoluteDateTime(), o1, o2)


class RelativeDateTime(Concatenable):
    FIELDS = ["years", "months", "weeks", "days", "hours", "minutes", "seconds"]

    def __init__(self, years=0, months=0, weeks=0, days=0, hours=0, minutes=0, seconds=0):
        self.years = years
        self.months = months
        self.weeks = weeks
        self.days = days
        self.hours = hours
        self.minutes = minutes
        self.seconds = seconds

    @classmethod
    def concatenate(cls, o1: 'RelativeDateTime', o2: 'RelativeDateTime') -> 'RelativeDateTime':
        return cls._concatenate(RelativeDateTime(), o1, o2)


class Constant(Printable):
    FIELDS = ["name", "alias", "value", "time_value"]

    def __init__(self, name, alias=None, value=None, options: List['ConstantOption'] = None, time_value: Callable = None):
        self.name = name
        self.alias = alias or []
        self.value = value
        self.options = options or []
        self.time_value = time_value

    def get_all(self):
        return [self.name, *self.alias]


class MethodEnum:
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return f"<Method: {self.name}>"

    def __repr__(self):
        return f"<Method: {self.name}>"



import datetime
from typing import Union

from datetimeparser.evaluator import Evaluator
from datetimeparser.parser import Parser


def parse(datetime_string: str, timezone: str = "Europe/Berlin") -> Union[datetime.datetime, int, None]:
    """
    Parses a datetime string and returns a datetime object.
    -1 is returned if the result is Infinity.
    If the datetime string cannot be parsed, None is returned.

    :param datetime_string: The datetime string to parse.
    :param timezone: The timezone to use. Should be a valid timezone for pytz.timezone(). Default: Europe/Berlin
    :return: A datetime object or an integer or None
    """
    parser_result = Parser(datetime_string).parse()

    if parser_result is None:
        return None

    evaluator_result = Evaluator(parser_result, tz=timezone).evaluate()

    if evaluator_result is None:
        return None

    return evaluator_result



from datetimeparser import parser
from datetimeparser import evaluator
from datetimeparser import enums
from datetimeparser import baseclasses
from datetimeparser import parsermethods

from datetimeparser.datetimeparser import parse



import re
from typing import Optional, Tuple, Union

from .enums import *
from .baseclasses import *


class RelativeDatetimeHelper:
    @staticmethod
    def from_keyword(keyword: Constant, delta: int = 1) -> RelativeDateTime:
        if keyword == DatetimeConstants.DAYS:
            return RelativeDateTime(days=delta)
        elif keyword == DatetimeConstants.WEEKS:
            return RelativeDateTime(weeks=delta)
        elif keyword == DatetimeConstants.MONTHS:
            return RelativeDateTime(months=delta)
        elif keyword == DatetimeConstants.YEARS:
            return RelativeDateTime(years=delta)
        elif keyword == DatetimeConstants.HOURS:
            return RelativeDateTime(hours=delta)
        elif keyword == DatetimeConstants.MINUTES:
            return RelativeDateTime(minutes=delta)
        elif keyword == DatetimeConstants.SECONDS:
            return RelativeDateTime(seconds=delta)


class AbsoluteDateFormatsParser:
    DATETIME_FORMATS = (
        "%Y.%m.%d %H:%M:%S",
        "%d.%m.%Y %H:%M:%S",
        "%Y.%m.%d",
        "%d.%m.%Y",
        "%H:%M:%S",
        "%H:%M",
        "%Y.%m.d %H:%M",
        "%d.%m.%Y %H:%M"
    )

    CLOCKTIME_FORMATS = (
        "%H:%M:%S",
        "%H:%M"
    )

    def parse(self, string: str) -> Union[None, Tuple[MethodEnum, AbsoluteDateTime]]:
        """
        Parses strings like "2020.01.01 12:00:00" or "2020.01.01"
        Returns None if the string cannot be parsed

        :param string: The string to parse
        :return: A tuple containing the method and the parsed data or None
        """
        for datetime_format in self.DATETIME_FORMATS:
            # Trying to find the right datetime format
            # And then instantly return the parsed datetime
            try:
                time = datetime.strptime(string, datetime_format)
            except ValueError:
                continue

            resulting_datetime = AbsoluteDateTime(hour=time.hour, minute=time.minute, second=time.second)

            if datetime_format not in self.CLOCKTIME_FORMATS:
                resulting_datetime.year = time.year
                resulting_datetime.month = time.month
                resulting_datetime.day = time.day

            return Method.ABSOLUTE_DATE_FORMATS, resulting_datetime

        return None


class RelativeDatetimesParser:
    PREPOSITIONS = ("in", "for", "next", "last")
    SKIPPABLE_KEYWORDS = ("and", "in", "for", ",")

    def parse(self, string: str) -> Union[None, Tuple[MethodEnum, RelativeDateTime]]:  # noqa: C901
        """
        Parses strings like "in 5 days" or "in 5 days and 3 hours"
        Returns None if the string cannot be parsed

        :param string: The string to parse
        :return: A tuple containing the method and the parsed data or None
        """

        # Checks if the string starts with a preposition
        # And then cuts it off because we need the preposition to differentiate future and past events
        for preposition in self.PREPOSITIONS:
            if string.startswith(preposition):
                string = string[len(preposition):]
                break
        else:
            preposition = ""

        new_data = []

        for argument in string.split():
            not_possible = True
            argument = argument.lower()

            # Skip unnecessary keywords
            if argument.lower() in self.SKIPPABLE_KEYWORDS:
                continue

            # an `a` always represents a 1 (e.g. a day = 1 day)
            elif argument.lower() == "a":
                new_data.append(1 if preposition != "last" else -1)
                not_possible = False

            # Cut off commas, they get in the way with parsing
            elif argument.endswith(","):
                argument = argument[:-1]

            # If the argument is a mini date format (m = minutes, s = seconds, ...) add the corresponding keyword to the list
            # Prepends a 1 if there wasn't a number set before the keyword (e.g. "s" doesnt make any sense so we prepend it with a 1)
            if len(argument) == 1:
                keyword = DatetimeConstants.convert_from_mini_date(argument)

                if keyword is not None:
                    if not new_data:
                        # If there is no number before the keyword, we need to add one ourselves
                        new_data.append(1 if preposition != "last" else -1)
                    elif not isinstance(new_data[-1], int):
                        # If the element before the keyword is not a number we have to add one ourselves
                        # This occurs when TODO
                        new_data.append(1 if preposition != "last" else -1)

                    new_data.append(keyword)
                    not_possible = False

            # If the argument is a number, simply add it to the list
            if argument.isnumeric():
                new_data.append(int(argument) if preposition != "last" else -int(argument))
                not_possible = False

            # If the argument represents a number (e.g. "one", "two", ...) simply add their integer value to the list
            for kw in NumberConstants.ALL:
                if argument.lower() in kw.get_all():
                    new_data.append(kw.value)
                    not_possible = False
                    break

            # If the argument is a datetime keyword (e.g. "year", "month", ...) add the keyword to the list
            # Prepends a 1 if there wasn't a number set before the keyword (e.g. "months" doesnt make any sense so we prepend it with a 1)
            for keyword in DatetimeConstants.ALL:
                if argument.lower() in keyword.get_all():
                    if not new_data:
                        # If there is no number before the keyword, we need to add one ourselves
                        new_data.append(1 if preposition != "last" else -1)
                    elif not isinstance(new_data[-1], int):
                        # If the element before the keyword is not a number we have to add one ourselves
                        # This occurs when TODO
                        new_data.append(1 if preposition != "last" else -1)

                    new_data.append(keyword)
                    not_possible = False
                    break
            else:
                # If the everything up to the last letter is numeric and the last letter is a valid mini date keyword (e.g. "30d", "10m", ...)
                # Add the number to the list and add the keyword to the list
                if argument[:-1].isnumeric():
                    number = int(argument[:-1]) if preposition != "last" else -int(argument[:-1])
                    keyword = argument[-1]

                    keyword = DatetimeConstants.convert_from_mini_date(keyword)

                    if keyword is not None:
                        new_data.append(number)
                        new_data.append(keyword)

                        not_possible = False

            # If none of the above if cases could find any data
            # Then we might've just encountered an invalid string
            # And other parsing methods might be able to parse it
            if not_possible:
                return None

        date = RelativeDateTime()

        # Has to be a multiple of 2 (because there's always a number before a keyword e.g. "1 day")
        # Otherwise it's not a valid relative date
        if len(new_data) % 2 != 0:
            return None

        while new_data:
            number = new_data.pop(0)
            type_ = new_data.pop(0)

            if type_ in DatetimeConstants.DATE:
                if type_ == DatetimeConstants.DAYS:
                    date.days = number
                elif type_ == DatetimeConstants.WEEKS:
                    date.weeks = number
                elif type_ == DatetimeConstants.MONTHS:
                    date.months = number
                elif type_ == DatetimeConstants.YEARS:
                    date.years = number
            elif type_ in DatetimeConstants.TIME:
                if type_ == DatetimeConstants.SECONDS:
                    date.seconds = number
                elif type_ == DatetimeConstants.MINUTES:
                    date.minutes = number
                elif type_ == DatetimeConstants.HOURS:
                    date.hours = number

        return Method.RELATIVE_DATETIMES, date


class ConstantsParser:
    # Weekday must go before DatetimeConstants because `friday` contains the word `day`
    CONSTANT_KEYWORDS = (*DatetimeDeltaConstants.ALL, *Constants.ALL, *MonthConstants.ALL, *WeekdayConstants.ALL, *DatetimeConstants.ALL)
    PREPOSITIONS = ("last", "next", "this", "previous", "at")
    PAST_PREPOSITIONS = ("last", "previous")
    FUTURE_PREPOSITIONS = ("next", "this")

    # Order is important because "at" and "the" are both in "at the"
    CUTOFF_KEYWORDS = ("at the", "in the", "at", "the")

    def _find_constant(self, argument: str) -> Optional[Constant]:
        """
        Finds a constant without any preposition or years
        Simply just "today" or "christmas"

        :param argument: The argument to look for
        :return: The constant if found, None otherwise
        """

        for keyword in self.CONSTANT_KEYWORDS:
            if argument in keyword.get_all():
                return keyword

        return None

    def parse(self, string: str) -> Optional[Tuple[MethodEnum, Tuple]]:  # noqa: C901
        """
        Parses strings like "today" or "next christmas" or "christmas 2022" or "last christmas"
        Returns None if the string cannot be parsed

        :param string: The string to parse
        :return: A tuple containing the method and the data or None
        """
        string = string.lower()

        # Cut off 'at', 'at the' and 'the'
        for cutoff_word in self.CUTOFF_KEYWORDS:
            if string.startswith(cutoff_word):
                string = string[len(cutoff_word):]

        # Strip whitespace
        string = " ".join(string.split())

        # Cut off the preposition if the strings starts with one and save the preposition
        # To differentiate future and past
        for preposition in self.PREPOSITIONS:
            if string.startswith(preposition):
                string = string[len(preposition):]
                break
        else:
            preposition = None

        string = " ".join(string.split())

        arguments = string.split()

        # If the last argument is a number, it must be a year
        # e.g. "christmas 2024"
        if arguments[-1].isnumeric():
            year = int(arguments.pop(-1))
            constant = self._find_constant(" ".join(arguments))

            if constant is None:
                return None
            else:
                return Method.CONSTANTS, (constant, AbsoluteDateTime(year=year))
        else:
            # Otherwise search for constants with or without prepositions
            # e.g. "christmas", "tomorrow", "tuesday"
            constant = self._find_constant(string)

            if constant is None:
                return None
            else:
                if preposition is None:
                    return Method.CONSTANTS, (constant,)
                else:
                    # Depending on the preposition the constant is either a future or past constant
                    if preposition in self.PAST_PREPOSITIONS:
                        return Method.CONSTANTS, (constant, RelativeDateTime(days=-1))
                    elif preposition in self.FUTURE_PREPOSITIONS:
                        if constant in DatetimeConstants.ALL:
                            if constant in DatetimeConstants.TIME:
                                return Method.CONSTANTS, (RelativeDatetimeHelper.from_keyword(constant, delta=1),)
                            elif constant in DatetimeConstants.DATE:
                                return Method.CONSTANTS, (RelativeDatetimeHelper.from_keyword(constant, delta=1),)
                        elif constant in [*WeekdayConstants.ALL, *Constants.ALL]:
                            return Method.CONSTANTS, (constant,)
                        else:
                            return None

        return None


class ConstantRelativeExtensionsParser:
    ABSOLUTE_KEYWORDS = (*Constants.ALL, *WeekdayConstants.ALL, *MonthConstants.ALL)
    RELATIVE_KEYWORDS = (*DatetimeDeltaConstants.ALL,)
    RELATIVE_TYPES = (RelativeDateTime,)
    CONSTANT_KEYWORDS = ABSOLUTE_KEYWORDS + RELATIVE_KEYWORDS

    PREPOSITIONS = ("last", "next", "this", "previous", "at", "on", "in")
    PAST_PREPOSITIONS = ("last", "previous")
    FUTURE_PREPOSITIONS = ("next", "this", "in")
    DATETIME_PREPOSITIONS = ("at", "on")

    # Order is important because "at" and "the" are both in "at the"
    CUTOFF_KEYWORDS = ("at the", "at", "the")

    @staticmethod
    def _find_number(string: str) -> Optional[int]:
        """
        Searches for a number in the string

        :param string: The string to search in
        :return: The number if found, None otherwise
        """

        if string == "a":
            return 1

        if string.isdecimal():
            return int(string)

        for constant in NumberConstants.ALL:
            for alias in constant.get_all():
                if string == alias:
                    return constant.value

        return None

    @staticmethod
    def _find_datetime_constant(string: str) -> Optional[Constant]:
        """
        Searches for a datetime constant in the string

        :param string: The string to search in
        :return: The constant if found, None otherwise
        """

        for constant in DatetimeConstants.ALL:
            for alias in constant.get_all():
                if string == alias:
                    return constant

        return None

    def _find_constant(self, argument: str) -> Optional[Constant]:
        """
        Finds a constant without any preposition or years
        Simply just "tomorrow" or "daylight change"

        :param argument: The argument to look for
        :return: The constant if found, None otherwise
        """

        for keyword in self.CONSTANT_KEYWORDS:
            if argument in keyword.get_all():
                return keyword

        return None

    def _get_preposition(self, string: str) -> Optional[Tuple[str, str]]:
        """
        Gets the preposition of a string if it exists (e.g. "next)
        :param string: the current string
        :return: the preposition if it exists otherwise None and the new string
        """
        string = " ".join(string.split())

        # Cut off the preposition if the strings starts with one and save the preposition
        # To differentiate future and past
        for preposition in self.PREPOSITIONS:
            if string.startswith(preposition):
                string = string[len(preposition):]
                break
        else:
            preposition = "at"

        string = " ".join(string.split())

        return preposition, string

    def _get_preposition_keyword(self, string: str) -> Optional[Tuple[str, str, Union[Constant, RelativeDateTime], str]]:
        """
        Gets the preposition and the constant of a string if it exists (e.g. "next daylight change)
        :param string: the current string
        :return: the preposition, the keyword as the string, the keyword and the new string if found otherwise None
        """
        # Cut off 'at', 'at the' and 'the'
        for cutoff_word in self.CUTOFF_KEYWORDS:
            if string.startswith(cutoff_word):
                string = string[len(cutoff_word):]
                string = string.strip()

        preposition, string = self._get_preposition(string)

        # "Bruteforce" the next arguments
        # Until a keyword has been found
        # The keyword can either be a constant, an absolute time, a clock time or a constant with a number
        # Unfortunately, we have to implement such parsing methods ourselves because we cannot use other parsers
        arguments = string.split()

        for i in range(1, len(arguments) + 1):
            tryable_keyword = " ".join(arguments[:i])

            # Try general keywords (e.g. "afternoon", "tomorrow", ...)
            # TODO: Add support for years (e.g. "monday 2021")
            keyword = self._find_constant(tryable_keyword)

            if keyword is not None:
                break

            # Try datetime delta constants (e.g. "17pm", "17:30", ...)
            datetime_delta_constants_parser = DatetimeDeltaConstantsParser()
            keyword = datetime_delta_constants_parser.parse(tryable_keyword)

            if keyword is not None:
                keyword = keyword[1]

                if preposition in self.DATETIME_PREPOSITIONS:
                    break
                else:
                    continue

            # Try absolute datetime formats (e.g. "17:30", "01.02.2020", ...)
            absolute_datetime_parser = AbsoluteDateFormatsParser()
            absolute_datetime_parser.DATETIME_FORMATS = (
                "%d.%m.%Y",
                "%Y.%m.%d"
            )
            absolute_datetime_parser.CLOCKTIME_FORMATS = ()
            keyword = absolute_datetime_parser.parse(tryable_keyword)

            if keyword is not None:
                keyword = keyword[1]

                if preposition in self.DATETIME_PREPOSITIONS:
                    break
                else:
                    return None

            # Try a mix of numbers and constants (e.g. "two weeks", "two years", ...)
            # TODO: This is the second only case where years could also be included
            # e.g. "(monday in) two weeks 2020"
            if len(tryable_keyword.split()) >= 2:
                number = self._find_number(tryable_keyword.split()[0])

                if number is None:
                    continue

                datetime_constant = self._find_datetime_constant(" ".join(tryable_keyword.split()[1:]))

                if datetime_constant is None:
                    continue

                keyword = RelativeDatetimeHelper.from_keyword(datetime_constant, delta=number)
                break
        else:
            return None

        return preposition, tryable_keyword, keyword, string

    def parse(self, string: str) -> Optional[Tuple[Method, Tuple]]:
        """
        Parses strings like "tomorrow afternoon" or "daylight change yesterday" or "monday in two weeks" or "tomorrow at 17"
        Returns None if the string cannot be parsed

        :param string: The string to parse
        :return: A tuple containing the method and the data or None
        """
        string = string.lower()
        result = self._get_preposition_keyword(string)

        if result is None:
            return None

        first_preposition, tryable_keyword, first_keyword, string = result
        string = string[len(tryable_keyword):]
        string = " ".join(string.split())

        # There has to be some contents left
        # Otherwise its not valid
        if not string:
            return None

        result = self._get_preposition_keyword(string)

        if result is None:
            return None

        second_preposition, tryable_keyword, second_keyword, string = result

        if first_keyword in self.ABSOLUTE_KEYWORDS and second_keyword in self.ABSOLUTE_KEYWORDS:
            return None

        if first_keyword in self.RELATIVE_KEYWORDS and second_keyword in self.RELATIVE_KEYWORDS:
            return None

        if first_keyword in self.RELATIVE_TYPES and second_keyword in self.RELATIVE_TYPES:
            return None

        if first_keyword in self.RELATIVE_KEYWORDS or type(second_keyword) in self.RELATIVE_TYPES:
            relative = first_preposition, first_keyword
            absolute = second_preposition, second_keyword
        else:
            relative = second_preposition, second_keyword
            absolute = first_preposition, first_keyword

        return Method.CONSTANTS_RELATIVE_EXTENSIONS, (*relative, *absolute)


class DatetimeDeltaConstantsParser:
    DATETIME_DELTA_CONSTANTS_PATTERN = re.compile("(([0-9]{1,2}:[0-9]{1,2})|([0-9]{1,2}))(am|pm|h| o\'clock|)")
    CLOCKTIME_FORMATS = (
        "%H:%M:%S",
        "%H:%M"
    )

    def parse(self, string: str) -> Optional[Tuple[MethodEnum, RelativeDateTime]]:  # noqa: C901
        """
        Parses strings like "at 3pm tomorrow" or "at 1am" or "at 10:30" or "at 17"
        Returns None if the string cannot be parsed

        :param string: The string to parse
        :return: A tuple containing the method and the data or None
        """

        data = string.split()

        if data[0] == "at":
            data.pop(0)

        if not data:
            return None

        # Matches the time and optional am/pm
        if self.DATETIME_DELTA_CONSTANTS_PATTERN.match(data[0]):
            time = data.pop(0)
            after_midday = time.endswith("pm") if time.endswith(("pm", "am")) else None
            time = time.replace("pm", "").replace("am", "").replace("h", "").replace(" o'clock", "")
            parsed_time = None

            for clocktime_format in self.CLOCKTIME_FORMATS:
                try:
                    parsed_time = datetime.strptime(time, clocktime_format)
                except ValueError:
                    continue

            # If the time does not match a clocktime format, does not contain a colon and is a number
            # e.g. "3(pm|am)", return that time respecting the after_midday flag
            if not parsed_time and time.count(":") == 0 and time.isdigit():
                if after_midday is not None:
                    parsed_time = RelativeDateTime(hours=(12 if after_midday else 0) + int(time))
                else:
                    parsed_time = RelativeDateTime(hours=int(time))
            elif parsed_time:
                if after_midday is not None:
                    parsed_time = RelativeDateTime(hours=(12 if after_midday else 0) + parsed_time.hour, minutes=parsed_time.minute, seconds=parsed_time.second)
                else:
                    parsed_time = RelativeDateTime(hours=parsed_time.hour, minutes=parsed_time.minute, seconds=parsed_time.second)
            else:
                return None

            # If the time is invalid return None
            if parsed_time.hours > 23 or parsed_time.minutes > 59 or parsed_time.seconds > 59:
                return None

            # If there's no more content left
            # Return the parsed time
            if not data:
                return Method.DATETIME_DELTA_CONSTANTS, parsed_time
            else:
                # Otherwise search for constants like
                # "at 3 in the morning"
                # "at 3 in the afternoon"

                rest = " ".join(data)

                # Find the matching keyword
                for kw in DatetimeDeltaConstants.ALL:
                    for alias in kw.get_all():
                        if rest in ("in the " + alias, alias):
                            value = kw.value
                            break
                    else:
                        continue

                    break
                else:
                    return None

                # "at 3pm in the morning" contains pm and in the morning
                # Is not valid so we return None
                if after_midday is not None:
                    return None

                # Value is either 0 or 12 (depending on the constant)
                parsed_time.hours += value

                # If the time is invalid return None
                if parsed_time.hours > 23:
                    return None

                return Method.DATETIME_DELTA_CONSTANTS, parsed_time

        return None


class AbsolutePrepositionParser:
    ABSOLUTE_PREPOSITION_TOKENS = (
        Keywords.OF,
        Keywords.AFTER,
        Keywords.BEFORE
    )

    RELATIVE_DATA_SKIPPABLE_WORDS = (
        "and",
        ",",
        "the"
    )

    RELATIVE_PAST_PREPOSITIONS = (
        "before"
    )

    MAX_RECURSION_DEPTH = 13

    def _split_data(self, string: str) -> Optional[Tuple[dict, dict, dict]]:
        """
        Splits the data into 3 parts
        The relative part, the keyword and the absolute part
        e.g. "3 months after christmas" -> ("3 months", "after", "christmas")

        :param string: The string to split
        :return: The splitted data or None
        """

        # If none of the known prepositions are in the string, return None
        if not any(absolute_preposition.name in string for absolute_preposition in self.ABSOLUTE_PREPOSITION_TOKENS):
            return None

        word = None
        char_count = 0

        # Count the number of characters up to the preposition in the string
        for word in string.split():
            char_count += len(word) + 1

            if word.lower() in map(lambda absolute_preposition: absolute_preposition.name, self.ABSOLUTE_PREPOSITION_TOKENS):
                break

        # Split the string into the relative part, the keyword and the absolute part
        # Subtract the length of the last word (because it's the preposition) and 2 because we accounted for 2 spaces in the loop
        relative = string[:char_count - len(word) - 2]
        absolute = string[char_count:]

        # There may be more prepositions in the absolute part so we try it again via recursion
        recursion_result = self._split_data(absolute)

        if recursion_result is not None:
            absolute = recursion_result

        return {'type': 'relative', 'data': relative}, {'type': 'keyword', 'data': word}, {'type': 'absolute', 'data': absolute}

    def _parse_relative_statement(self, relative_statement: str) -> List[Union[int, Constant]]:  # noqa: C901
        """
        Parses strings like "3 seconds, 2 minutes and 4 hours", "the fifth day", "4. week"
        It resembles `relative_datetimes`

        :param relative_statement: The statement to parse
        :return: List of integers and keywords
        """

        arguments = relative_statement.lower().split()
        returned_data = []

        for argument in arguments:
            if argument.endswith(","):
                argument = argument[:-1]

            if argument.startswith(","):
                argument = argument[1:]

            # Skip words like "and", "," or "the"
            if argument in self.RELATIVE_DATA_SKIPPABLE_WORDS:
                continue

            # 'a' means the same as '1' (e.g. 'a day' => '1 day')
            if argument == "a":
                returned_data.append(1)
                continue

            if argument.isnumeric():
                returned_data.append(int(argument))
                continue

            found_keyword = False

            # '1st', '1.', 'first', ...
            for keyword in NumberCountConstants.ALL:
                for alias in keyword.get_all():
                    if alias == argument:
                        returned_data.append(keyword.value)
                        found_keyword = True
                        break
                else:
                    continue

                break

            if found_keyword:
                continue

            # 'one', 'two', 'three', ...
            for keyword in NumberConstants.ALL:
                for alias in keyword.get_all():
                    if alias == argument:
                        returned_data.append(keyword.value)
                        found_keyword = True
                        break
                else:
                    continue

                break

            if found_keyword:
                continue

            # 'seconds', 'minutes', 'hours', ...
            for keyword in DatetimeConstants.ALL:
                for alias in keyword.get_all():
                    if alias == argument:
                        returned_data.append(keyword)
                        break
                else:
                    continue

                break

        return returned_data

    def _concatenate_relative_data(self, relative_data_tokens: List[Union[int, Constant]], preposition: str) -> List[Union[int, Constant, RelativeDateTime]]:
        """
        Concatenates [1, RelativeDate(DAY), 2, RelativeDate(MONTH)] into [RelativeDate(days=1, months=2)]
        respecting the preposition (future and past)

        :param relative_data_tokens: The already parsed relative data
        :param preposition: The preposition
        :return: List of Keywords and integers
        """

        # If the data does not divide by 2, we cannot concatenate it
        # e.g. "1st (of july)"
        if len(relative_data_tokens) % 2 != 0:
            return relative_data_tokens

        returned_data = []

        for i in range(0, len(relative_data_tokens), 2):
            value, unit = relative_data_tokens[i], relative_data_tokens[i + 1]

            # We can only concatenate Datetime Constants like 'seconds', 'days', 'months', ...
            if isinstance(value, int) and unit in DatetimeConstants.ALL:
                if preposition in self.RELATIVE_PAST_PREPOSITIONS:
                    value *= -1

                if unit in DatetimeConstants.DATE:
                    current_data = RelativeDatetimeHelper.from_keyword(unit, value)
                elif unit in DatetimeConstants.TIME:
                    current_data = RelativeDatetimeHelper.from_keyword(unit, value)
                else:
                    raise RuntimeError("Unknown Datetime Constant:", unit)

                # We cannot concatenate anything if there's no data
                if i == 0:
                    returned_data.append(current_data)
                else:
                    # Check if the latest entry is the same type as the current unit
                    if isinstance(returned_data[-1], type(current_data)):
                        data_before = returned_data.pop()
                        new_data = data_before.concatenate(data_before, current_data)
                        returned_data.append(new_data)
                    else:
                        # Otherwise we cannot concatenate both parts of the data so we just append the current one
                        returned_data.append(current_data)

        return returned_data

    def _parse_absolute_statement(self, data: Union[str, Tuple]) -> Optional:
        """
        Parses constant values like "christmas", "previous st. patricks day", ..
        If the input is a Tuple containing the absolute preposition tokens, we return the result of`_convert_tokens`
        This happens because strings like "(10 days after) 3 months before christmas" are also valid

        :param data: Either the string to parse or a Tuple containing absolute preposition tokens
        :return: The parsed result (no typing specified because there are just too many values)
        """

        if isinstance(data, str):
            # TODO: Call ConstantRelativeExtensionsParser as well
            constants_parser = ConstantsParser()
            constants_parser.CONSTANT_KEYWORDS = (*Constants.ALL, *MonthConstants.ALL, *WeekdayConstants.ALL)
            constants_parser.PREPOSITIONS = ("last", "next", "this", "previous")
            constants_parser.PAST_PREPOSITIONS = ("last", "previous")
            constants_parser.FUTURE_PREPOSITIONS = ("next", "this")
            constants_parser.CUTOFF_KEYWORDS = ("the",)

            result = constants_parser.parse(data)

            if result is None:
                # If the result is None there may be just a normal year (e.g. "2018")
                if data.isnumeric():
                    return (AbsoluteDateTime(year=int(data)),)
                else:
                    return None
            else:
                # The first element is the Method signature (Method.CONSTANTS)
                return result[1]
        else:
            return self._convert_tokens(data)

    def _convert_tokens(self, tokens: Tuple[dict, dict, dict]) -> Optional[List[Union[int, Constant, RelativeDateTime]]]:
        """
        Converts and parses the splitted tokens from _split_data into Tokens suited for the Evaluator

        :param tokens: Tokens from _split_data
        :return: A List containing numbers and keywords
        """

        returned_data = []

        # Choose the right parsing method for the specific data types
        for i, data_part in enumerate(tokens):
            if data_part["type"] == "relative":
                # "3 hours and 4 minutes (after christmas)", "the 5th (of july)", ...
                relative_data = []

                for relative_part_data in self._parse_relative_statement(data_part["data"]):
                    relative_data.append(relative_part_data)

                preposition = tokens[1]["data"]
                relative_data = self._concatenate_relative_data(relative_data, preposition)

                for relative_part_data in relative_data:
                    returned_data.append(relative_part_data)
            elif data_part["type"] == "keyword":
                # "of", "before", ...
                for keyword in self.ABSOLUTE_PREPOSITION_TOKENS:
                    if keyword.name == data_part["data"]:
                        returned_data.append(keyword)
                        break
                else:
                    raise ValueError("No preposition given")
            elif data_part["type"] == "absolute":
                absolute_data = self._parse_absolute_statement(data_part["data"])

                if absolute_data is None:
                    return None

                for absolute_part_data in absolute_data:
                    returned_data.append(absolute_part_data)

        return returned_data

    def parse(self, string: str) -> Optional[Tuple[MethodEnum, List[Union[int, Constant, RelativeDateTime]]]]:
        """
        Parses strings like "3 days after christmas" or "1 hour, 2 minutes and 5 days after 3 months before christmas"
        Returns None if the string cannot be parsed

        :param string: The string to parse
        :return: A tuple containing the method and the data or None
        """

        prepared_data = self._split_data(string)

        if prepared_data is None:
            return None

        data = self._convert_tokens(prepared_data)

        if data is None:
            return None

        return Method.ABSOLUTE_PREPOSITIONS, data



from datetime import datetime, timedelta


def eastern_calc(year_time: int) -> datetime:
    a = year_time % 19
    k = year_time // 100
    m = 15 + (3 * k + 3) // 4 - (8 * k + 13) // 25
    d = (19 * a + m) % 30
    s = 2 - (3 * k + 3) // 4
    r = d // 29 + (d // 28 - d // 29) * (a // 11)
    og = 21 + d + r
    sz = 7 - (year_time + year_time // 4 + s) % 7
    oe = 7 - (og - sz) % 7
    os = og + oe

    if os > 32:
        return datetime(year=year_time, month=4, day=(os - 31))
    else:
        return datetime(year=year_time, month=3, day=os)


def thanksgiving_calc(year_time: int) -> datetime:
    year_out = datetime(year=year_time, month=11, day=29)
    date_out = datetime(year=year_time, month=11, day=3)
    return year_out - timedelta(days=(date_out.weekday() + 2))


def days_feb(year_time: int) -> int:
    if int(year_time) % 400 == 0 or int(year_time) % 4 == 0 and not int(year_time) % 100 == 0:
        return 29
    else:
        return 28


def year_start(year_time: int) -> datetime:
    return datetime(year=year_time, month=1, day=1)



from typing import Union

from .baseclasses import *
from .enums import *


class AbsolutePrepositionsEvaluator:

    @staticmethod
    def sanitize_input(parsed_list: list) -> list:
        """removes useless keywords"""
        return [element for element in parsed_list if element not in Keywords.ALL and not isinstance(element, str)]

    @staticmethod
    def get_base(sanitized_input: list, year: int) -> datetime:
        if isinstance(sanitized_input[-1], AbsoluteDateTime):
            if isinstance(sanitized_input[-2], Constant):
                if isinstance(sanitized_input[-3], int):
                    dt: datetime = sanitized_input[-2].time_value(sanitized_input[-1].year)
                    day: int = sanitized_input[-3]
                    return datetime(dt.year, dt.month, day, dt.hour, dt.minute, dt.second)
                return sanitized_input[-2].time_value(sanitized_input[-1].year)
            return datetime(sanitized_input[-1].year, 1, 1)
        elif isinstance(sanitized_input[-1], Constant):
            if isinstance(sanitized_input[-2], int):
                dt: datetime = sanitized_input[-1].time_value(year)
                day: int = sanitized_input[-2]
                return datetime(dt.year, dt.month, day, dt.hour, dt.minute, dt.second)
            return sanitized_input[-1].time_value(year)

    @staticmethod
    def calc_relative_time(sanitized_list: list) -> RelativeDateTime:
        rel_list = [relative_time for relative_time in sanitized_list if isinstance(relative_time, RelativeDateTime)]

        ev_out = RelativeDateTime()
        for i in rel_list:
            ev_out.years += i.years
            ev_out.months += i.months
            ev_out.weeks += i.weeks
            ev_out.days += i.days
            ev_out.hours += i.hours
            ev_out.minutes += i.minutes
            ev_out.seconds += i.seconds

        return ev_out

    @staticmethod
    def remove_milli_seconds(dt: datetime) -> datetime:
        return datetime.strptime(dt.strftime("%Y-%m-%d %H:%M:%S"), "%Y-%m-%d %H:%M:%S")


def evaluate_absolute_date_formats(current_time: datetime, parsed: AbsoluteDateTime) -> AbsoluteDateTime:
    ev_out = AbsoluteDateTime(
        year=current_time.year if parsed.year == 0 else parsed.year,
        month=current_time.month if parsed.month == 0 else parsed.month,
        day=current_time.day if parsed.day == 0 else parsed.day,
        hour=parsed.hour,
        minute=parsed.minute,
        second=parsed.second
    )

    return ev_out


def evaluate_constant_relatives(current_time: datetime, parsed: list) -> datetime:
    sanitized = AbsolutePrepositionsEvaluator.sanitize_input(parsed)
    base: datetime = current_time
    ev_out = None

    if isinstance(sanitized[-1], Constant):
        base = sanitized[-1].time_value(current_time.year)
        hour, minute, sec = sanitized[-2].time_value(None)
        ev_out = datetime(base.year, base.month, base.day, hour, minute, sec)

    elif isinstance(sanitized[-1], RelativeDateTime):
        base += relativedelta(
            years=sanitized[-1].years,
            months=sanitized[-1].months,
            weeks=sanitized[-1].weeks,
            days=sanitized[-1].days,
            hours=sanitized[-1].hours,
            minutes=sanitized[-1].minutes,
            seconds=sanitized[-1].seconds
        )
        if sanitized[-2] in WeekdayConstants.ALL:
            ev_out = datetime.strptime(
                sanitized[-2].time_value(base),
                "%Y-%m-%d %H:%M:%S"
            )
        elif sanitized[-2] in Constants.ALL_RELATIVE_CONSTANTS:
            base = sanitized[-2].time_value(None)
            hour, minute, sec = sanitized[-1].hours, sanitized[-1].minutes, sanitized[-1].seconds
            ev_out = datetime(base.year, base.month, base.day, hour, minute, sec)

        elif sanitized[-2] in DatetimeDeltaConstants.ALL:
            ev_out = datetime(
                year=base.year,
                month=base.month,
                day=base.day,
                hour=sanitized[-2].time_value(None)[0],
                minute=sanitized[-2].time_value(None)[1],
                second=sanitized[-2].time_value(None)[2]
            )

    elif isinstance(sanitized[-1], AbsoluteDateTime):
        base = datetime(
            year=current_time.year if sanitized[-1].year == 0 else sanitized[-1].year,
            month=current_time.month if sanitized[-1].month == 0 else sanitized[-1].month,
            day=current_time.day if sanitized[-1].day == 0 else sanitized[-1].day,
            hour=sanitized[-1].hour,
            minute=sanitized[-1].minute,
            second=sanitized[-1].second
        )
        hour, minute, sec = sanitized[-2].time_value(None)
        ev_out = datetime(base.year, base.month, base.day, hour, minute, sec)

    return ev_out


def evaluate_absolute_prepositions(current_time: datetime, parsed: list) -> datetime:
    ev = AbsolutePrepositionsEvaluator
    base_year = current_time.year
    sanitized = ev.sanitize_input(parsed)
    base = ev.get_base(sanitized, base_year)
    rel_out = ev.calc_relative_time(sanitized)

    base += relativedelta(
        years=rel_out.years,
        months=rel_out.months,
        weeks=rel_out.weeks,
        days=rel_out.days,
        hours=rel_out.hours,
        minutes=rel_out.minutes,
        seconds=rel_out.seconds
    )

    return ev.remove_milli_seconds(base)


def evaluate_constants(current_time: datetime, parsed_object) -> Union[AbsoluteDateTime, datetime, int]:
    dt: datetime = current_time
    object_type: Constant = parsed_object[0]

    if len(parsed_object) == 2:
        if isinstance(parsed_object[0], Constant):
            object_year: AbsoluteDateTime = parsed_object[1].year
            dt = object_type.time_value(object_year)

            if current_time > dt and object_year == 0:
                dt += relativedelta(years=1)

    else:
        if object_type.name == "infinity":
            return object_type.value

        elif object_type in WeekdayConstants.ALL:
            dt: datetime = datetime.strptime(
                object_type.time_value(current_time),
                "%Y-%m-%d %H:%M:%S"
            )

        else:
            dt = object_type.time_value(current_time.year)

            if isinstance(dt, tuple):
                dt = datetime(
                    year=current_time.year,
                    month=current_time.month,
                    day=current_time.day,
                    hour=dt[0],
                    minute=dt[1],
                    second=dt[2]
                )

        if current_time > dt and parsed_object[0] not in Constants.ALL_RELATIVE_CONSTANTS:
            dt += relativedelta(years=1)

    ev_out = AbsoluteDateTime(
        dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second
    )

    return ev_out


def evaluate_relative_datetime(current_time: datetime, parsed: RelativeDateTime) -> AbsoluteDateTime:
    out: datetime = current_time

    out += relativedelta(
        years=parsed.years,
        months=parsed.months,
        weeks=parsed.weeks,
        days=parsed.days,
        hours=parsed.hours,
        minutes=parsed.minutes,
        seconds=parsed.seconds
    )

    ev_out = AbsoluteDateTime(
        out.year, out.month, out.day, out.hour, out.minute, out.second
    )

    return ev_out


def evaluate_datetime_delta_constants(current_time: datetime, parsed: RelativeDateTime) -> AbsoluteDateTime:
    ev_out = AbsoluteDateTime(
        current_time.year, current_time.month, current_time.day, parsed.hours, parsed.minutes, parsed.seconds
    )

    return ev_out



import string as string_utils

from .parsermethods import (
    AbsoluteDateFormatsParser,
    RelativeDatetimesParser,
    ConstantsParser,
    ConstantRelativeExtensionsParser,
    DatetimeDeltaConstantsParser,
    AbsolutePrepositionParser
)


class Parser:
    PROCEDURE = (
        AbsoluteDateFormatsParser().parse,
        RelativeDatetimesParser().parse,
        ConstantsParser().parse,
        ConstantRelativeExtensionsParser().parse,
        DatetimeDeltaConstantsParser().parse,
        AbsolutePrepositionParser().parse
    )

    def __init__(self, string):
        self.string = self.remove_non_ascii(" ".join(string.split()))

    @staticmethod
    def remove_non_ascii(string):
        return ''.join(c for c in string if c in string_utils.printable)

    def parse(self):
        result = None

        for method in self.PROCEDURE:
            if result is None:
                result = method(self.string)
            else:
                break

        return result



from pytz import timezone, UnknownTimeZoneError

from .evaluatormethods import *


class Evaluator:
    def __init__(self, parsed_object, tz="Europe/Berlin"):
        try:
            tiz = timezone(tz)
        except UnknownTimeZoneError:
            raise ValueError("Unknown timezone: {}".format(tz))

        self.parsed_object_type = parsed_object[0]
        self.parsed_object_content: Union[list, AbsoluteDateTime, RelativeDateTime] = parsed_object[1]
        self.current_datetime: datetime = datetime.strptime(datetime.strftime(datetime.now(tz=tiz), "%Y-%m-%d %H:%M:%S"), "%Y-%m-%d %H:%M:%S")

    def evaluate(self) -> Union[datetime, int, None]:
        ev_out = None

        if self.parsed_object_type == Method.ABSOLUTE_DATE_FORMATS:
            ev_out = evaluate_absolute_date_formats(
                self.current_datetime,
                self.parsed_object_content
            )

        if self.parsed_object_type == Method.ABSOLUTE_PREPOSITIONS:
            ev_out = evaluate_absolute_prepositions(
                self.current_datetime,
                self.parsed_object_content
            )

        if self.parsed_object_type == Method.CONSTANTS:
            ev_out = evaluate_constants(
                self.current_datetime,
                self.parsed_object_content
            )

            if not isinstance(ev_out, AbsoluteDateTime):
                return ev_out

        if self.parsed_object_type == Method.RELATIVE_DATETIMES:
            ev_out = evaluate_relative_datetime(
                self.current_datetime,
                self.parsed_object_content
            )

        if self.parsed_object_type == Method.CONSTANTS_RELATIVE_EXTENSIONS:
            ev_out = evaluate_constant_relatives(
                self.current_datetime,
                self.parsed_object_content
            )

        if self.parsed_object_type == Method.DATETIME_DELTA_CONSTANTS:
            ev_out = evaluate_datetime_delta_constants(
                self.current_datetime,
                self.parsed_object_content
            )

        if isinstance(ev_out, datetime):
            return ev_out
        try:
            dt_object = datetime(
                ev_out.year,
                ev_out.month,
                ev_out.day,
                ev_out.hour,
                ev_out.minute,
                ev_out.second
            )
            return dt_object
        except ValueError:
            return None



from distutils.core import setup

# for the readme
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding="utf-8")

setup(
  name='python-datetimeparser',
  long_description_content_type="text/markdown",
  long_description=long_description,
  packages=['datetimeparser'],
  version='0.4rc',  # version number: https://peps.python.org/pep-0440/
  license='MIT',
  description='A parser library built for parsing the english language into datetime objects.',
  author='Ari24',
  author_email='ari.publicmail@gmail.com',
  url='https://github.com/aridevelopment-de/datetimeparser',
  download_url='https://github.com/aridevelopment-de/datetimeparser/archive/v0.1e.tar.gz',
  keywords=['datetime', 'parser', 'parsing', 'grammar', 'datetime-parser'],
  install_requires=[
          'python-dateutil',
          'pytz',
          'typing'
      ],
  classifiers=[
    'Development Status :: 4 - Beta',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3.7',      #Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10'
  ],
)



from datetime import datetime, timedelta
from enum import Enum, auto

from dateutil.relativedelta import relativedelta

from .baseclasses import Constant, MethodEnum
from .formulars import days_feb, eastern_calc, thanksgiving_calc, year_start


class ConstantOption(Enum):
    TIME_VARIABLE = auto()
    DATE_VARIABLE = auto()
    YEAR_VARIABLE = auto()


class Constants:
    CHRISTMAS = Constant('christmas', ['xmas'], options=[ConstantOption.YEAR_VARIABLE],
                         time_value=lambda year_time: datetime(year=year_time, month=12, day=25))
    SILVESTER = Constant('silvester', ['new years eve'], options=[ConstantOption.YEAR_VARIABLE],
                         time_value=lambda year_time: datetime(year=year_time, month=12, day=31))
    EASTERN = Constant('eastern', ['easter'], options=[ConstantOption.YEAR_VARIABLE], time_value=eastern_calc)
    NICHOLAS = Constant('nicholas', ['nicholas day'], options=[ConstantOption.YEAR_VARIABLE],
                        time_value=lambda year_time: datetime(year=year_time, month=12, day=6))
    HALLOWEEN = Constant('halloween', options=[ConstantOption.YEAR_VARIABLE],
                         time_value=lambda year_time: datetime(year=year_time, month=10, day=31))
    APRIL_FOOLS_DAY = Constant('april fools day', ['april fool day'], options=[ConstantOption.YEAR_VARIABLE],
                               time_value=lambda year_time: datetime(year=year_time, month=4, day=1))
    THANKSGIVING = Constant('thanksgiving', options=[ConstantOption.YEAR_VARIABLE], time_value=thanksgiving_calc)
    SAINT_PATRICKS_DAY = Constant('saint patrick\'s day',
                                  ['saint patricks day', 'st. patrick\'s day', 'saint pt. day', 'st patrick\'s day', 'st patricks day'],
                                  options=[ConstantOption.YEAR_VARIABLE],
                                  time_value=lambda year_time: datetime(year=year_time, month=3, day=17))
    VALENTINES_DAY = Constant('valentines day', ['valentine', 'valentine day'], options=[ConstantOption.YEAR_VARIABLE],
                              time_value=lambda year_time: datetime(year=year_time, month=2, day=14))

    SUMMER_BEGIN = Constant('summer begin', ['summer', 'begin of summer', 'begin of the summer'], options=[ConstantOption.YEAR_VARIABLE],
                            time_value=lambda year_time: datetime(year=year_time, month=6, day=1))
    WINTER_BEGIN = Constant('winter begin', ['winter', 'begin of winter', 'begin of the winter'], options=[ConstantOption.YEAR_VARIABLE],
                            time_value=lambda year_time: datetime(year=year_time, month=12, day=1))
    SPRING_BEGIN = Constant('spring begin', ['spring', 'begin of spring', 'begin of the spring'], options=[ConstantOption.YEAR_VARIABLE],
                            time_value=lambda year_time: datetime(year=year_time, month=3, day=1))
    FALL_BEGIN = Constant('fall begin',
                          ['fall', 'begin of fall', 'begin of the fall', 'autumn begin', 'autumn', 'begin of autumn', 'begin of the autumn'],
                          options=[ConstantOption.YEAR_VARIABLE], time_value=lambda year_time: datetime(year=year_time, month=9, day=1))
    SUMMER_END = Constant('summer end', ['end of summer', 'end of the summer'], options=[ConstantOption.YEAR_VARIABLE],
                          time_value=lambda year_time: datetime(year=year_time, month=8, day=31, hour=23, minute=59, second=59))
    WINTER_END = Constant('winter end', ['end of winter', 'end of the winter'], options=[ConstantOption.YEAR_VARIABLE],
                          time_value=lambda year_time: datetime(year=year_time, month=2, day=days_feb(year_time), hour=23, minute=59, second=59))
    SPRING_END = Constant('spring end', ['end of spring', 'end of the spring'], options=[ConstantOption.YEAR_VARIABLE],
                          time_value=lambda year_time: datetime(year=year_time, month=5, day=31, hour=23, minute=59, second=59))
    FALL_END = Constant('fall end', ['end of fall', 'end of the fall', 'autumn end', 'end of autumn', 'end of the autumn'],
                        options=[ConstantOption.YEAR_VARIABLE],
                        time_value=lambda year_time: datetime(year=year_time, month=11, day=30, hour=23, minute=59, second=59))

    MORNING = Constant('morning', ['at morning'],
                       options=[ConstantOption.YEAR_VARIABLE, ConstantOption.DATE_VARIABLE])
    EVENING = Constant('evening', ['at evening'],
                       options=[ConstantOption.YEAR_VARIABLE, ConstantOption.DATE_VARIABLE])
    LUNCHTIME = Constant('lunchtime', ['lunch'], options=[ConstantOption.YEAR_VARIABLE, ConstantOption.DATE_VARIABLE])

    BEGIN_AOC = Constant('aoc begin', ['aoc', 'begin of aoc', 'begin of the aoc', 'advent of code begin', 'advent of code', 'begin of advent of code', 'begin of the advent of code'],
                         options=[ConstantOption.YEAR_VARIABLE], time_value=lambda year_time: datetime(year=year_time, month=12, day=1, hour=6))
    END_AOC = Constant('aoc end', ['end of aoc', 'end of the aoc', 'advent of code end', 'end of advent of code', 'end of the advent of code'],
                       options=[ConstantOption.YEAR_VARIABLE], time_value=lambda year_time: datetime(year=year_time, month=12, day=26, hour=6))

    END_OF_YEAR = Constant('end of year', ['the end of year', 'the end of the year', 'end of the year'],
                           options=[ConstantOption.YEAR_VARIABLE],
                           time_value=lambda year_time: datetime(year=year_time, month=12, day=31, hour=23, minute=59, second=59))
    BEGIN_OF_YEAR = Constant('begin of year', ['the begin of year', 'the begin of the year', 'begin of the year'],
                             options=[ConstantOption.YEAR_VARIABLE], time_value=year_start)

    INFINITY = Constant('infinity', ['inf'], value=-1)

    TODAY = Constant('today', options=[ConstantOption.TIME_VARIABLE], time_value=lambda _: datetime(datetime.today().year, datetime.today().month, datetime.today().day))
    TOMORROW = Constant('tomorrow', options=[ConstantOption.TIME_VARIABLE], time_value=lambda _: datetime.today() + relativedelta(days=1))
    YESTERDAY = Constant('yesterday', options=[ConstantOption.TIME_VARIABLE], time_value=lambda _: datetime.today() - relativedelta(days=1))
    NOW = Constant('now', ['at the moment', 'current time', 'current time now'], time_value=lambda _: datetime.now())

    ALL = [
        CHRISTMAS, SILVESTER, EASTERN, NICHOLAS, HALLOWEEN, APRIL_FOOLS_DAY, THANKSGIVING, SAINT_PATRICKS_DAY, VALENTINES_DAY,
        SUMMER_END, WINTER_END, SPRING_END, FALL_END, SUMMER_BEGIN, WINTER_BEGIN, SPRING_BEGIN, FALL_BEGIN,
        MORNING, EVENING, LUNCHTIME,
        BEGIN_AOC, END_AOC,
        END_OF_YEAR, BEGIN_OF_YEAR,
        INFINITY,
        TODAY, TOMORROW, YESTERDAY, NOW
    ]
    ALL_RELATIVE_CONSTANTS = [TODAY, TOMORROW, YESTERDAY, NOW]


class DatetimeDeltaConstants:
    # time_value is a tuple containing (hours, minutes, seconds)
    MIDNIGHT = Constant('midnight', value=0, options=[ConstantOption.DATE_VARIABLE], time_value=lambda _: (0, 0, 0))
    NIGHT = Constant('night', value=0, options=[ConstantOption.DATE_VARIABLE], time_value=lambda _: (21, 0, 0))
    MORNING_NIGHT = Constant('morning night', value=0, options=[ConstantOption.DATE_VARIABLE],
                             time_value=lambda _: (3, 0, 0))
    DAYLIGHT_CHANGE = Constant('daylight change', ['daylight saving', 'daylight saving time'], value=0,
                               options=[ConstantOption.YEAR_VARIABLE, ConstantOption.DATE_VARIABLE],
                               time_value=lambda _: (6, 0, 0))
    SUNRISE = Constant('sunrise', value=0, options=[ConstantOption.DATE_VARIABLE], time_value=lambda _: (7, 0, 0))
    MORNING = Constant('morning', value=0, options=[ConstantOption.DATE_VARIABLE], time_value=lambda _: (6, 0, 0))
    BREAKFAST = Constant('breakfast', value=0, options=[ConstantOption.DATE_VARIABLE], time_value=lambda _: (8, 0, 0))

    MIDDAY = Constant('midday', value=12, options=[ConstantOption.DATE_VARIABLE], time_value=lambda _: (12, 0, 0))
    LUNCH = Constant('lunch', ['lunchtime'], value=12, options=[ConstantOption.DATE_VARIABLE],
                     time_value=lambda _: (12, 0, 0))
    AFTERNOON = Constant('afternoon', value=12, options=[ConstantOption.DATE_VARIABLE], time_value=lambda _: (15, 0, 0))
    EVENING = Constant('evening', value=12, options=[ConstantOption.DATE_VARIABLE], time_value=lambda _: (18, 0, 0))
    DINNER = Constant('dinner', ['dinnertime'], value=12, options=[ConstantOption.DATE_VARIABLE],
                      time_value=lambda _: (19, 0, 0))
    DAWN = Constant('dawn', value=12, options=[ConstantOption.DATE_VARIABLE], time_value=lambda _: (6, 0, 0))
    DUSK = Constant('dusk', value=12, options=[ConstantOption.DATE_VARIABLE], time_value=lambda _: (20, 0, 0))
    SUNSET = Constant('sunset', value=12, options=[ConstantOption.DATE_VARIABLE], time_value=lambda _: (18, 30, 0))

    ALL = [
        MORNING, AFTERNOON, EVENING, NIGHT, MORNING_NIGHT, DAYLIGHT_CHANGE, MIDNIGHT, MIDDAY, DAWN, DUSK,
        SUNRISE, SUNSET, LUNCH, DINNER, BREAKFAST
    ]


class NumberConstants:
    # Presented to you by github copilot
    ONE = Constant('one', value=1)
    TWO = Constant('two', value=2)
    THREE = Constant('three', value=3)
    FOUR = Constant('four', value=4)
    FIVE = Constant('five', value=5)
    SIX = Constant('six', value=6)
    SEVEN = Constant('seven', value=7)
    EIGHT = Constant('eight', value=8)
    NINE = Constant('nine', value=9)
    TEN = Constant('ten', value=10)
    ELEVEN = Constant('eleven', value=11)
    TWELVE = Constant('twelve', value=12)
    THIRTEEN = Constant('thirteen', value=13)
    FOURTEEN = Constant('fourteen', value=14)
    FIFTEEN = Constant('fifteen', value=15)
    SIXTEEN = Constant('sixteen', value=16)
    SEVENTEEN = Constant('seventeen', value=17)
    EIGHTEEN = Constant('eighteen', value=18)
    NINETEEN = Constant('nineteen', value=19)
    TWENTY = Constant('twenty', value=20)
    TWENTY_ONE = Constant('twenty one', alias=["twentyone", "twenty-one"], value=21)
    TWENTY_TWO = Constant('twenty two', alias=["twentytwo", "twenty-two"], value=22)
    TWENTY_THREE = Constant('twenty three', alias=["twentythree", "twenty-three"], value=23)
    TWENTY_FOUR = Constant('twenty four', alias=["twentyfour", "twenty-four"], value=24)
    TWENTY_FIVE = Constant('twenty five', alias=["twentyfive", "twenty-five"], value=25)
    TWENTY_SIX = Constant('twenty six', alias=["twentysix", "twenty-six"], value=26)
    TWENTY_SEVEN = Constant('twenty seven', alias=["twentyseven", "twenty-seven"], value=27)
    TWENTY_EIGHT = Constant('twenty eight', alias=["twentyeight", "twenty-eight"], value=28)
    TWENTY_NINE = Constant('twenty nine', alias=["twentynine", "twenty-nine"], value=29)
    THIRTY = Constant('thirty', value=30)
    THIRTY_ONE = Constant('thirty one', alias=["thirtyone", "thirty-one"], value=31)

    # Reversed to avoid conflicts with other constants (one is included in twenty one)
    ALL = [ONE, TWO, THREE, FOUR, FIVE, SIX, SEVEN, EIGHT, NINE, TEN,
           ELEVEN, TWELVE, THIRTEEN, FOURTEEN, FIFTEEN, SIXTEEN, SEVENTEEN, EIGHTEEN, NINETEEN, TWENTY,
           TWENTY_ONE, TWENTY_TWO, TWENTY_THREE, TWENTY_FOUR, TWENTY_FIVE, TWENTY_SIX, TWENTY_SEVEN, TWENTY_EIGHT, TWENTY_NINE,
           THIRTY, THIRTY_ONE][::-1]


class NumberCountConstants:
    # Presented to you by github copilot
    FIRST = Constant('first', alias=['1st', '1.'], value=1)
    SECOND = Constant('second', alias=['2nd', '2.'], value=2)
    THIRD = Constant('third', alias=['3rd', '3.'], value=3)
    FOURTH = Constant('fourth', alias=['4th', '4.'], value=4)
    FIFTH = Constant('fifth', alias=['5th', '5.'], value=5)
    SIXTH = Constant('sixth', alias=['6th', '6.'], value=6)
    SEVENTH = Constant('seventh', alias=['7th', '7.'], value=7)
    EIGHTH = Constant('eighth', alias=['8th', '8.'], value=8)
    NINTH = Constant('ninth', alias=['9th', '9.'], value=9)
    TENTH = Constant('tenth', alias=['10th', '10.'], value=10)
    ELEVENTH = Constant('eleventh', alias=['11th', '11.'], value=11)
    TWELFTH = Constant('twelfth', alias=['12th', '12.'], value=12)
    THIRTEENTH = Constant('thirteenth', alias=['13th', '13.'], value=13)
    FOURTEENTH = Constant('fourteenth', alias=['14th', '14.'], value=14)
    FIFTEENTH = Constant('fifteenth', alias=['15th', '15.'], value=15)
    SIXTEENTH = Constant('sixteenth', alias=['16th', '16.'], value=16)
    SEVENTEENTH = Constant('seventeenth', alias=['17th', '17.'], value=17)
    EIGHTEENTH = Constant('eighteenth', alias=['18th', '18.'], value=18)
    NINETEENTH = Constant('nineteenth', alias=['19th', '19.'], value=19)
    TWENTIETH = Constant('twentieth', alias=['20th', '20.'], value=20)
    TWENTY_FIRST = Constant('twenty first', alias=['21st', '21.', 'twentyfirst', 'twenty-first'], value=21)
    TWENTY_SECOND = Constant('twenty second', alias=['22nd', '22.', 'twentysecond', 'twenty-second'], value=22)
    TWENTY_THIRD = Constant('twenty third', alias=['23rd', '23.', 'twentythird', 'twenty-third'], value=23)
    TWENTY_FOURTH = Constant('twenty fourth', alias=['24th', '24.', 'twentyfourth', 'twenty-fourth'], value=24)
    TWENTY_FIFTH = Constant('twenty fifth', alias=['25th', '25.', 'twentyfifth', 'twenty-fifth'], value=25)
    TWENTY_SIXTH = Constant('twenty sixth', alias=['26th', '26.', 'twentysixth', 'twenty-sixth'], value=26)
    TWENTY_SEVENTH = Constant('twenty seventh', alias=['27th', '27.', 'twentyseventh', 'twenty-seventh'], value=27)
    TWENTY_EIGHTH = Constant('twenty eighth', alias=['28th', '28.', 'twentyeighth', 'twenty-eighth'], value=28)
    TWENTY_NINTH = Constant('twenty ninth', alias=['29th', '29.', 'twentyninth', 'twenty-ninth'], value=29)
    THIRTIETH = Constant('thirtieth', alias=['30th', '30.'], value=30)
    THIRTY_FIRST = Constant('thirty first', alias=['31st', '31.', 'thirthyfirst', "thirty-first"], value=31)

    # Reversed to avoid conflicts with other constants
    ALL = [FIRST, SECOND, THIRD, FOURTH, FIFTH, SIXTH, SEVENTH, EIGHTH, NINTH, TENTH,
           ELEVENTH, TWELFTH, THIRTEENTH, FOURTEENTH, FIFTEENTH, SIXTEENTH, SEVENTEENTH, EIGHTEENTH, NINETEENTH, TWENTIETH,
           TWENTY_FIRST, TWENTY_SECOND, TWENTY_THIRD, TWENTY_FOURTH, TWENTY_FIFTH, TWENTY_SIXTH, TWENTY_SEVENTH, TWENTY_EIGHTH, TWENTY_NINTH,
           THIRTIETH, THIRTY_FIRST][::-1]


class DatetimeConstants:
    SECONDS = Constant('seconds', ['second', 'sec', 'secs'])
    MINUTES = Constant('minutes', ['minute', 'min', 'mins'])
    HOURS = Constant('hours', ['hour'])
    DAYS = Constant('days', ['day'])
    WEEKS = Constant('weeks', ['week'])
    MONTHS = Constant('months', ['month'])
    YEARS = Constant('years', ['year'])

    TIME = [SECONDS, MINUTES, HOURS]
    DATE = [DAYS, WEEKS, MONTHS, YEARS]
    ALL = [SECONDS, MINUTES, HOURS, DAYS, WEEKS, MONTHS, YEARS]

    @classmethod
    def convert_from_mini_date(cls, md):
        if md.lower() == "s":
            return cls.SECONDS
        elif md.lower() == "m":
            return cls.MINUTES
        elif md.lower() == "h":
            return cls.HOURS
        elif md.lower() == "w":
            return cls.WEEKS
        elif md.lower() == "d":
            return cls.DAYS
        elif md.lower() == "y":
            return cls.YEARS


class WeekdayConstants:
    MONDAY = Constant('monday', time_value=lambda date: f"{date + timedelta((0 - date.weekday()) % 7)}")
    TUESDAY = Constant('tuesday', time_value=lambda date: f"{date + timedelta((1 - date.weekday()) % 7)}")
    WEDNESDAY = Constant('wednesday', time_value=lambda date: f"{date + timedelta((2 - date.weekday()) % 7)}")
    THURSDAY = Constant('thursday', time_value=lambda date: f"{date + timedelta((3 - date.weekday()) % 7)}")
    FRIDAY = Constant('friday', time_value=lambda date: f"{date + timedelta((4 - date.weekday()) % 7)}")
    SATURDAY = Constant('saturday', time_value=lambda date: f"{date + timedelta((5 - date.weekday()) % 7)}")
    SUNDAY = Constant('sunday', time_value=lambda date: f"{date + timedelta((6 - date.weekday()) % 7)}")

    ALL = [MONDAY, TUESDAY, WEDNESDAY, THURSDAY, FRIDAY, SATURDAY, SUNDAY]


class MonthConstants:
    JANUARY = Constant('january', ['jan'], time_value=lambda year_time: datetime(year=year_time, month=1, day=1))
    FEBRUARY = Constant('february', ['feb'], time_value=lambda year_time: datetime(year=year_time, month=2, day=1))
    MARCH = Constant('march', ['mar'], time_value=lambda year_time: datetime(year=year_time, month=3, day=1))
    APRIL = Constant('april', ['apr'], time_value=lambda year_time: datetime(year=year_time, month=4, day=1))
    MAY = Constant('may', time_value=lambda year_time: datetime(year=year_time, month=5, day=1))
    JUNE = Constant('june', ['jun'], time_value=lambda year_time: datetime(year=year_time, month=6, day=1))
    JULY = Constant('july', ['jul'], time_value=lambda year_time: datetime(year=year_time, month=7, day=1))
    AUGUST = Constant('august', ['aug'], time_value=lambda year_time: datetime(year=year_time, month=8, day=1))
    SEPTEMBER = Constant('september', ['sep'], time_value=lambda year_time: datetime(year=year_time, month=9, day=1))
    OCTOBER = Constant('october', ['oct'], time_value=lambda year_time: datetime(year=year_time, month=10, day=1))
    NOVEMBER = Constant('november', ['nov'], time_value=lambda year_time: datetime(year=year_time, month=11, day=1))
    DECEMBER = Constant('december', ['dec'], time_value=lambda year_time: datetime(year=year_time, month=12, day=1))

    ALL = [JANUARY, FEBRUARY, MARCH, APRIL, MAY, JUNE, JULY, AUGUST, SEPTEMBER, OCTOBER, NOVEMBER, DECEMBER]


class Keywords:
    OF = Constant('of')
    AFTER = Constant('after')
    BEFORE = Constant('before')
    NEXT = Constant('next')
    IN = Constant('in')
    FOR = Constant('for')

    ALL = [OF, AFTER, BEFORE, NEXT, IN, FOR]


class Method:
    ABSOLUTE_PREPOSITIONS = MethodEnum('absolute_prepositions')
    ABSOLUTE_DATE_FORMATS = MethodEnum('absolute_date_formats')
    CONSTANTS = MethodEnum('constants')
    CONSTANTS_RELATIVE_EXTENSIONS = MethodEnum('constants_relative_extensions')
    DATETIME_DELTA_CONSTANTS = MethodEnum('datetime_delta_constants')
    RELATIVE_DATETIMES = MethodEnum('relative_datetimes')
